#include <NucModel.h>
#include <Particle.h>
#include <PhysicsFunctions.h>
#include <MathFunctions.h>

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <sstream>
#include <iostream>
#include <sigmaA_disi.h>
#include <lossesN.h>
#include <lossesPI.h>
#include <IR_fast.h>

using namespace std;


NucModel::NucModel(eModelType model, int LossType, TRandom3* myRandom) :
  fnelements(0),
  fType(eNone),
  fRandom(myRandom)
{
  fLosses = LossType;
  fType = model;
  if(model==eOnlyProton) { //only proton 

    fnelements = 1;
    fe_1  = 30.; // Photodisintegration minimum energy (MeV) 
    fe_max=150.; // Photodisintegration maximum energy (MeV) 
    fZ.push_back(1);
    fA.push_back(1);

  }
  else if(model==eSalomon) {

    fnelements = 51;
    fe_1  =  e_1; // Photodisintegration minimum energy (MeV) 
    fe_max =  e_max; // Photodisintegration maximum energy (MeV)     

    for (int i=0; i<fnelements; i++) {
     
      fZ.push_back(ZZ[i]);
      fA.push_back(AA[i]);
      feth_1.push_back(eth_1[i]);
      feth_2.push_back(eth_2[i]);
      fe0_1.push_back(e0_1[i]);
      fcsi_1.push_back(csi_1[i]);
      fe0_2.push_back(e0_2[i]);
      fcsi_2.push_back(csi_2[i]);
      fdelta_2.push_back(delta_2[i]);
      fzita.push_back(zita[i]);
    
    }

  
  }
  else {
  
  }



}

NucModel::~NucModel() {
}


void NucModel::Validate(Particle* input) {

  if(fType==eOnlyProton) { //each particle is transformed into proton
    input->SetMassNum(1);
    input->SetCharge(1);
    input->SetType(eProton);
  }
  else if(fType==eSalomon) {

  }
}


int NucModel::TestPrimary(int mass) {

  if(fType==eOnlyProton && mass==1) { //each particle is transformed into proton
    return 1;
  }

  int found = binary(mass); 
  return found;

}


int NucModel::binary(int mass) { //return the charge of the element ... if found
    
  int first = 0;
  int last = fnelements;

  while(first<=last) {
    int mid = (first + last)/2;
    if(mass > fA[mid])
      last = mid - 1;
    else if (mass < fA[mid])
      first = mid + 1;
    else
      return fZ[mid];
  }

  return -1;
}


Particle* NucModel::GetRandomParticle(double E, double z) {
  
  Particle * output;
  const double index = fRandom->Integer(fnelements);

  const double A = fA[index];
  const double Z = fZ[index];
  
  if(A>1) output = new Particle(A, Z, 0, E, z, eNucleus);
  else output = new Particle(1, 1, 0, E, z, eProton);
  
  return output;
}


vector<Particle> NucModel::PropagateParticle(Particle* input) {

  vector<Particle> output;  
  if(fType==eOnlyProton && input->GetType() != eProton) { 
      cout << "Only protons can be propagated within this model!" << endl;
      exit(-1);
  }

  if(input->GetType()==eProton) output.push_back(*PropagateProton(input));
  else output = PropagateNucleus(input);
  
  return output;

}


Particle* NucModel::PropagateProton(Particle* input) {

  double Gamdet = evolveN(input->GetZprod(),input->GetGprod(),0); //protons always reach the Earth 
  
  Particle * output = input;
  input->SetZint(0);
  input->SetGint(Gamdet);
  input->SetBranch(input->GetBranch());
  
  return output; 
  
}

vector<Particle> NucModel::PropagateNucleus(Particle* input) {

  cout<<"... propagating nucleus " << input->GetMassNum() << " Ecurr "
      << input->GetEprod() << " from z " << input->GetZprod() << endl;

  vector<Particle> output = GetLosses(input);

  if ( output.size()>1 ) {
    cout<<"...... int type " << output.size()-1 << " " 	<< input->GetMassNum() 
    	<< "-->" << output[0].GetMassNum() << " Efin " << input->GetEint() 
     	<< endl;
    
  } 
  else {
    
    if (input->GetMassNum() > 1) {
      
      input->SetZint(0);
      double EEnd = GetEdetN(input);
      cout<<"........... nucleus "<< input->GetMassNum() 
	  <<" reaches Earth with Edet = " << EEnd 
	  << endl;
    }
  } 

 
  return output;

}


vector<Particle> NucModel::GetLosses(Particle * input)
{

   vector<Particle> output; 
   int Acurr = input->GetMassNum();
   int Afin = Acurr;
   int intType = 0;
   double dist = 0;
   double Efin = 0;
   double zfin = GetProcess(intType, dist, Efin, *input); 
   int Zfin;
    if (zfin == 0) intType = -1;
   
   if (intType > 0) {
     Afin =  Acurr - intType;
 
     const int nBr = input->GetBranch() + 1; 
     input->SetEint(Efin);
     input->SetZint(zfin);  
     if( Afin <= 1 ) { // nucleus fully fragmented
       Afin = 1;
       Zfin = 1;
       intType = Acurr;
       const double Eproton = input->GetGint()*mN;
       Particle* nucleon = new Particle(Afin,Zfin,nBr,Eproton,zfin,eProton); 
       cout << "Nucleus fully fragmented" << endl;
       for(int i=0; i<intType; i++) output.push_back(*nucleon);
     } 
     else {
       if( GetJNuc(Afin) < 0 ) {
	 int jnuc = -GetJNuc(Afin); 
	 Afin = fA[jnuc];
	 Zfin = fZ[jnuc];
	 intType = Acurr - Afin;
       }
       else {
	 Zfin = fZ[GetJNuc(Afin)];
       }
       const double Eproton = input->GetGint()*mN;
       const double Enucleo = input->GetGint()*mN*(double)Afin;
       Particle* nucleo;
       if(Afin == 4)  {
	 nucleo = new Particle(Afin,Zfin,nBr,Enucleo,zfin,eNucleus);
	 output.push_back(*nucleo);
	 if( Acurr == 9 ){
	   intType = Acurr - 2*Afin;
	   output.push_back(*nucleo);
	 }
	 cout<<"....... stacking p + 4He at z " << zfin <<endl;
       }
       else {
	 if(Afin>1) {
	   nucleo = new Particle(Afin,Zfin,nBr,Enucleo,zfin,eNucleus);
	   output.push_back(*nucleo);
	 }
	 else {
	   nucleo = new Particle(Afin,Zfin,nBr,Enucleo,zfin,eProton);
	   output.push_back(*nucleo);
	   cout<<"....... stacking 1 protons at z "<<zfin
	       <<" with Ep "<<Enucleo<<endl;
	 }
       }
       Particle* nucleon = new Particle(1,1,nBr,Eproton,zfin,eProton); 
       cout << "....... stacking " << intType << " protons at z " << zfin
	   <<" with Ep "<< Eproton << endl;
       for(int i=0; i<intType; i++) output.push_back(*nucleon); 
     }

   }
   input->SetIntMult(intType);

   return output; 

}

// -------------------------------------------------------------------------

int NucModel::GetJNuc(int ANuc)
{
  int jNuc;
  int i;
  if(ANuc > fA[0] || ANuc < 0) return 0;
  if(ANuc == 1) return -1;
  for(i = 0; i < fnelements; i++) if(fA[i] <= ANuc) break;
  if(fA[i] == ANuc) jNuc = i;
  else jNuc = -i;
  return jNuc;
}
// -------------------------------------------------------------------------



// -------------------------------------------------------------------------
double NucModel::GetEdetN(Particle *input)
{ 
  double Gamcurr = input->GetGprod();
  int Acurr = input->GetMassNum();
  double zcurr = input->GetZprod();
  double zfin = input->GetZint();
  double Gamdet = evolveA0(Acurr,zcurr,Gamcurr,zfin); 
  input->SetGint(Gamdet);

  return Gamdet*mN*(double)Acurr;
}


// -------------------------------------------------------------------------

double NucModel::GetProcess(int &ndis, double &distInt, double &EInt, Particle &input)
{
  double nAprop = 0;
  double Gamcurr = input.GetGprod(); 
  double zmin = 1e-8;
  double zS = input.GetZprod();
  int Acurr =  input.GetMassNum();

  double IntnA[1001];
  double zdis[1001];
  
  double zmax;
  double dz;
  double t3;

  int jcurr = GetJNuc (Acurr);

  for(int i=0; i<=1000; i++)
    {
      IntnA[i]=1; //afg
      zdis[i]=0;
    }
    
  t3=0.;
	    
  zmax = zS;
  dz=(log(zmax)-log(zmin))/nstep;
 
  double z0=zS;
  double z=zS;
  double t2in=0.;
  double t2=0.;
  double Gamold=Gamcurr;
  
  IntnA[0] = 0.;
  zdis[0] = zS;

  int lastj=0;
  double u = fRandom->Uniform(0., 1.);
  int ndisDummy = -1; // sp


  for (int j = 1; j <= nstep; j++) // db
    {
      z = zS-exp(log(zmin)+j*dz);
    
      double ddz=z-z0; 
      double dtdz = 1./ (H0 *(1.+z) * sqrt(om_m *pow(1+z,3) + om_l)); 
      double Gam_out =  evolveA0(Acurr,z0,Gamold,z); // afg
      Gamold=Gam_out; // afg

      double beta_disi = losseAdisi(Acurr,z,Gam_out,jcurr, ndisDummy); // sp 
	
      t2  = dtdz * 1 * beta_disi;
      t3+=ddz*(t2+t2in)/2.;	    
      nAprop = exp(t3);
	
      IntnA[j]=1.-nAprop; //afg
	
      zdis[j]=z;
      t2in=t2;
	
      z0=z; 

      lastj=j;

      if(IntnA[j]>=u) break;      
    }
 
  double a = 0.;
  double b = 0.;
  double zInt = 0.;
  double x = 0.;
  const double dzlast = zdis[lastj] - zdis[lastj-1];
 
  b = (IntnA[lastj]-IntnA[lastj-1])/dzlast;
  a = IntnA[lastj-1] - (zdis[lastj-1]*b);
  x = (u - a)/b;
   
  if (x<0.) x=0.;
  zInt=x;
 
  double Gam_int = evolveA0(Acurr,zS,Gamcurr,zInt);
  EInt = Gam_int*mN*Acurr;

  distInt = qgaus(rr,zInt,zS); 

  double bdis = SecInY*SpeedOfLight*m2cm/(Mpc2cm*losseAdisi(Acurr,zInt,Gam_int,jcurr,ndis));
  input.SetBetaDis(bdis);
  double zero = 0;
  const double dfin = GetR(zero, zInt);
  if(zInt==0.) ndis = 0;
  
  cout << "..... zInt = " << zInt << " dEarth " << dfin << " int.length(zInt) " << bdis << endl; 

  return zInt; 
}

// -------------------------------------------------------------------------


double NucModel::evolveA0 (int &A_0, double &z_in, double &Gam_in,
		 double &z_out)
{
//.............................................................................
// Evolution with red-shift of Gamma and A for nuclei,
// .............................................................................
  k_com = GetJNuc(A_0);

  double z1 = z_in;
  double z2 = z_out;
  const double eps=1.e-4; //afg
  double h1 =z1-z2;  //afg
  double hmin = 0.;

  int nok, nbad;

  double Lny0 = log(Gam_in);

  double Lny = odeint(Lny0, z1, z2, eps, h1, hmin, nok , nbad, derivsA0);

  double Gam_out = -999.;
  Gam_out = exp(Lny);
  return Gam_out;
}  



// -------------------------------------------------------------------------

double NucModel::losseAdisi(int &Acurr, double &z, double &Gam, int &k0, int &multi)
{
  //............................................................................. 
  // Subroutine that computes the photo-disintegration energy losses for nuclei 
  // expressed in 1/y. All backgrounds (CMB+IR/V/UV) are taken into account
  //.............................................................................
  double sum1=0;
  double sum2=0;
  double sum3=0;

  Gam_com = Gam;
  z_com = z;
  k_com = k0;
  s_com = 0;

  //.... photodisintegration energy losses (CMB + IRV/UV) ....

  double lxmin1 = log(eth_1[k0]);
  double lxmax1 = log(eth_2[k0]);
  sum1 = qgaus (fdisi, lxmin1, lxmax1);

  double lxmin2 = log(eth_2[k0]);
  double lxmax2 = log(e_1);
  sum2 = qgaus (fdisi, lxmin2, lxmax2);

  double lxmin3 = log(e_1);
  double lxmax3 = log(e_max);
  sum3 = qgaus (fdisi, lxmin3, lxmax3);
  
  double sum=sum1+sum2+sum3;

  // sp
  if (sum<=0) sum = 0;
  if (multi < 0) return sum;

  multi = 0; 
  // sp

  double urange = fRandom->Uniform(0., 1.);
  double umult  = fRandom->Uniform(0., 1.);

  int erange = 1;
  if(urange<=(sum1/(sum))) multi = 1;

  else if(urange>(sum1/sum)&&urange<=(sum1+sum2)/sum) {
    erange = 2;
    s_com = 1;
    double sum21 = qgaus (fdisi, lxmin2, lxmax2);
    multi = 2;
    if (umult < sum21/sum2) multi = 1;
  }	
  else {
    erange = 3;
    int dim = (int)BRdisi(Acurr, 0);
    double cum = 0;
    for (int n = 1; n <= dim; n++ ) {
      cum += BRdisi(Acurr, n);
      if (umult < cum) {
	multi = n;
	break;
      }
    }
    if (Acurr == 9) multi = 5;
  }

  double beta_disi = sum; 
//.........................................................................

  return beta_disi;
}


// -------------------------------------------------------------------------

double derivsA0 (double &z, double &Lny)
{
// Right hand side of the differential equations for energy losses


 double Gam = exp(Lny);      
 double dtdz = 1./ (H0 * (1+z) * sqrt(om_m * pow((1+z),3) + om_l)); 

 double beta_pair = losseA0pair(z,Gam,fA[k_com]);
 double beta_pi = losseA0Pi(z,Gam,fA[k_com]);

 return dtdz * beta_pair + 1/(1+z);
 
}

// -------------------------------------------------------------------------

double fdisi (double lx)
{
// integral function of the photo-disntegration energy losses

  double Ene = exp(lx);

  double sigma_1 = 0;
  double sigma_2 = 0;
  double sigma_3 = 0;
 
  sigma_disi (k_com, Ene, sigma_1, sigma_2, sigma_3);
  
  double sigma_tot = 0.;

  if(s_com == 0) sigma_tot = sigma_1 + sigma_2 + sigma_3;
  if(s_com == 1) sigma_tot = sigma_1;
  if(s_com == 2) sigma_tot = sigma_2;
  if(s_com == 3) sigma_tot = sigma_3;

  double In_IR = 0.;

//................. photodisintegration on CMB ............................
  double N0_CMB = 1.452e12;  // normalization CMB 1/y/MeV^2/mbarn
  double kT     = 2.327e-10; // KT (MeV)
  double xx1    = Ene / (2.*(1.+z_com) * Gam_com * kT);
  double In_CMB = pow(1+z_com,3) * N0_CMB * log(1./(1-exp(-xx1)));
//.........................................................................

//................. photodisintegration on IR/V/UV ........................

  if (fLosses == 1) {
    //.......... Malkan & Stecker (2006) ..................
    double N0_IR = 4.734e2 ; // normalization  1/y/MeV^2/mbarn
    double e0    = 1.e-6; // 1 eV (MeV)
    double xx2   = Ene / (2.* Gam_com * e0);
    double xmin  = 3.3e-3; // (2.d-3) minimum contributing IR/V/UV energy (eV)
    double xmax  = 1.e1; // (1.6d1) maximum contributing IR/V/UV energy (eV)
    double sum   = 0.;
    if (z_com >= 5) sum=0.;
    else {
      double lxmin, lxmax;
      if (xx2 <= xmax) {
	if(xx2 >= xmin) {
	  lxmin = log(xx2);
	  lxmax = log(xmax);
          sum = qgaus(fdisi_IR,lxmin,lxmax);
	} else {
	  lxmin = log(xmin);
	  lxmax = log(xmax);
          sum = qgaus(fdisi_IR,lxmin,lxmax);
	}
      }
    }
    In_IR = N0_IR * sum;
  }
  if (fLosses == 2) {
//.....................................................

//................. simple power law ..................

    double gam_IR = 2.5; // IR/V/UV background power law index 
    double N0_IR = 2.*5.207e-2/(gam_IR+1); // (min) normalization 1/y/MeV^2/mbarn
    double e0 = 1.e-6; // 1 eV (MeV)
    double xx2 = Ene/(2*Gam_com*e0);
    double xmin = 2.e-3; // (3.e-3) minimum contributing IR/V/UV energy (eV)
    double xmax=1.; // (13.6)  maximum contributing IR/V/UV energy (eV)
    
    double sum = 0;
    if (z_com < 5.) {
      if (xx2 <= xmax) {
	if (xx2 >= xmin) sum = pow(xx2, -(gam_IR+1));
	else sum= pow(2.e-3,-(gam_IR+1));
      }
    }
    if (z_com <= 1.4) In_IR = N0_IR * pow(1+z_com, 3.1) * sum;
    else In_IR = N0_IR * sum;
  }
  double ln = In_IR + In_CMB / ((1.+z_com) * (1.+z_com));
  double retvalue = Ene * sigma_tot * Ene * ln / (1. * Gam_com * Gam_com); // 1/y

  return retvalue;
}

// -------------------------------------------------------------------------
double NucModel::BRdisi(int A, int imult)
{
  const int Alow[4] = {1, 9, 10, 23};
  const int Aup[4]  = {4, 9, 22, 56};
  const int dim[4]  = {2, 1,  6, 15};

  const double fBR[] = {0.8, 0.2,                          // He
			1.0,                               // Be
			0.1, 0.3, 0.1, 0.1, 0.2, 0.2,      // B
			0.1, 0.35, 0.1, 0.05, 0.15,        // Na
			0.045, 0.040, 0.035, 0.030, 0.025,
			0.020, 0.018, 0.015, 0.012, 0.010};
  int indx = -1;
  int ptr  = 0;
  for (int i = 0; i < 4; i++) {
    if (A>=Alow[i] && A<=Aup[i]) {
      indx = i;
      break;
    }
    ptr += dim[i];
  }

  int mult = 0;
  if (indx >= 0) mult = dim[indx];

  if (imult <= 0) return (double)mult;
  else return fBR[ptr+imult-1];
}

// -------------------------------------------------------------------------


void sigma_disi (int &k, double &Ene, 
		 double &sigma_1, double &sigma_2, double &sigma_3)
{
// ............................................................................. 
//  Subroutine that computes the photodisintegration cross-section,
//  requires (Z,A) and photon energy (MeV) in input. The output is in mbarn 
//  and is divided into sigma_1: single nucleon emission (g,n) and (g,p); 
//  sigma_2: double nucleon emission (g,2n), (g,2p) and (g,np); 
//  sigma_3: constant normalized cross-section. We have followed the recipes 
//  of Stecker and Salamon ApJ 512 (1999) 521.
// .............................................................................

  sigma_1 = 0.;
  sigma_2 = 0.;
  sigma_3 = 0.;
  if (Ene >= e_max) return;

  int A  = fA[k];
  int Ze = fZ[k];

  //...... Normalization factor (mb*MeV) ..................
  double Sigma_d = 60.* (double)((A-Ze)*Ze)/ (double)(A); // mb*MeV
  
  if ((Ene >= eth_1[k]) && (Ene <= e_1)) {
    double W1 = W(eth_1[k], e0_1[k], delta_1[k]);
    sigma_1 = (csi_1[k] * Sigma_d/W1) *
      exp(-2.*((Ene-e0_1[k])/delta_1[k])*((Ene-e0_1[k])/delta_1[k]));
  } else sigma_1 = 0.;

  if ((Ene >= eth_2[k]) && (Ene <= e_1)) {
    double W2 = W(eth_2[k], e0_2[k], delta_2[k]);
    sigma_2 = (csi_2[k] * Sigma_d/W2 )*
      exp(-2.*((Ene-e0_2[k])/delta_2[k])*((Ene-e0_2[k])/delta_2[k]));
  } else sigma_2=0.;

  if ((Ene >= e_1) && (Ene <=e_max))
    {
    sigma_3 = zita[k] * Sigma_d/(e_max-e_1);
    }
  else sigma_3=0.;
}
// -------------------------------------------------------------------------


double losseA0pair(double &z, double &Gam, const int &A)
{
  double * y2a = new double[nE_ee];

//............................................................................. 
// Subroutine that computes the pair production energy losses for nuclei 
// expressed in 1/y. Only the CMB background is relevant
//.............................................................................

//........ pair production energy losses (only CMB) ......................

  double E = mN * Gam * (1+z); // corresponding proton energy (eV)
  double beta_pair = 0.;
  double lbeta;
  if (E <= 1.e17) beta_pair = 0.;
  else {
    if(E <= 1.e24){
      spline(lEne_ee,lbeta0_ee,nE_ee,1.e30,1.e30,y2a);
      lbeta = splint(lEne_ee,lbeta0_ee,y2a,nE_ee,log10(E));
    } 
    else lbeta = -11.91234 - 0.61714286 * (log10(E)-24.);
    beta_pair = pow(1+z,3) * ((double)A/4) * pow(10, lbeta); // 1/y (Z/A=1/2)
  }
  delete [] y2a;
  return beta_pair;
}

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
double losseA0Pi(double &z, double &Gam, const int &A)
{
  double * y2a = new double[n_pi];

//............................................................................. 
// Subroutine that computes the photopion production energy losses for nuclei 
// expressed in 1/y. Only the CMB background is relevant
//.............................................................................

//........ photopion production energy losses (only CMB) ......................

  double G = Gam * (1+z); // corresponding proton energy (eV)
  double lambda = 0.;
  double lambda_A = 0.;
  double beta_pi = 0.;
  if (log10(G) <= 9.8) beta_pi = 0.;
  else {
    if(log10(G) <= 13.19){
      spline(G_pi,lambda_pi,n_pi,1.e30,1.e30,y2a);
      lambda = splint(G_pi,lambda_pi,y2a,n_pi,log10(G));
      lambda_A = pow(10.,lambda) / A; // Mpc
    }
    else lambda_A = 5.76 / A; // Mpc
 
    beta_pi = pow(1+z,3) * SecInY*SpeedOfLight*m2cm / (Mpc2cm * lambda_A); //1/yr 
  }
  delete [] y2a;
  return beta_pi;
}

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------

double fdisi_IR (double lx)
{
  // integral function for the IR energy losses
  
  double x = exp(lx);
  double lxx = log10(x);
  double * y2a = new double[nE_IR*nz_IR];

  splie2(lE_IR,z_IR,ln_IR,nE_IR,nz_IR,y2a);
  double lnn = splin2(lE_IR,z_IR,ln_IR,y2a,nE_IR,nz_IR,lxx,z_com);
  
  delete [] y2a;

  double ln=pow(10.0,lnn)/x;
  double f_IR=x*(ln)/x/x;
  
  return f_IR;

}

// -------------------------------------------------------------------------
