//
// File   : SimProp.cc
//
// Purpose: Program to simulate propagation of nuclei
//
// Authors: D. Boncioli, A. Grillo, S. Petrera, F. Salamida
//
// Based on the analitical calculation by R. Aloisio 
//      (arXiv:1006.2484 [astro-ph.CO] and arXiv:0802.4452 [astro-ph])
// 
// date: 10/09/2010
// Last update: 25/10/2011
 
int VERSION = 2;
int REVISION = 0;

#include <iomanip>  
#include <vector>
#include <cstdio>
#include <unistd.h>
#include <iostream>
#include <fstream>
#include <sstream>
#include <cstdlib>
#include <list>

#include <cmath>
#include <time.h>

//ROOT headers
#include <TROOT.h>
#include <TRandom3.h>
#include <TFile.h>
#include <TTree.h>
#include <TBranch.h>
TRandom3* myRandom;

#include <Particle.h>
#include <NucModel.h>
#include <Output.h> 

clock_t start, end=0;
double cpu_time_used;
 

int  Nevts = 100;
int  Areq     = 56;
int  LossType = 1;
unsigned int  Seed = 65539;
double  zmin = 0.0;
double  zmax = 1.0;
double  Emin = 1.e17;
double  Emax = 1.e21;
int NatEarth[56] = {0};
int event = 0;

using namespace std;

vector<Particle> Follow(vector<Particle>, NucModel*, Output*);

bool GetOptions(int, char **);

bool PrintSummary(int);


int main(int argc,char*argv[])
{


  if (!GetOptions(argc, argv))
    return 0;
 
  cout << '\n';
  cout << "=====================================================================\n";
  cout << "SimProp v" << VERSION << "r" << REVISION ;

  cout << '\n';
 
  bool random = false;
  if (Areq == 0) random = true;
  
  bool proton = false;
  if (Areq == 1) proton = true;

  myRandom = new TRandom3();  
  myRandom->SetSeed(Seed);
  
  //NucModel* model = new NucModel(eOnlyProton,LossType,myRandom);

  NucModel* model = new NucModel(eSalomon,LossType,myRandom);
  int Zreq = 0;
  if (!proton && Areq > 0) {
    Zreq = model->TestPrimary(Areq);
    if(Zreq<0) {
      cout<<"** Initial nucleus not in the list"<<endl;
      exit(EXIT_FAILURE);
    }
  }
  if (Nevts <= 0) {
    cout<<"** Wrong choice for requested number of events N"<<endl;
    exit(EXIT_FAILURE);
  }
  if (Areq < 0) {
    cout<<"** Wrong choice for requested nucleus A"<<endl;
    exit(EXIT_FAILURE);
  }
  if(Emin>Emax) {
    cout<<"** Wrong choice for requested energy range"<<endl;
    exit(EXIT_FAILURE);
  }
  if (LossType<0 || LossType>2) {
    cout<<"** Wrong choice for requested Energy Loss model L"<<endl;
    exit(EXIT_FAILURE);
  }
  if (zmin*zmax< 0 || zmin >zmax) {
    cout<<"** Wrong choice for requested redshift range"<<endl;
    exit(EXIT_FAILURE);
  }

  cout << "** Initial nucleus is: " << Areq << endl;
  cout << "** Losses are: ";
  if (LossType == 0) cout << " WARNING: events are simulated with only CMB"; 
  else{  
    cout << " CMB";
    if (LossType == 1) cout << " + IR (Malkan & Stecker)"; 
    if (LossType == 2) cout << " + IR (power law)"; 
  }
  cout<<endl;
  cout<<"** Limits on redshift are: "<<zmin<<" "<<zmax<<endl; 
  cout<<"** Maximum energy is: "<<Emax<<" eV"<<endl;

  char FileOut[150];
  sprintf(FileOut,"SimProp-v%dr%d_N%d_A%d_L%d_z%4.2f_Z%4.2f_e%4.1f_E%4.1f_s%d.root", VERSION, REVISION, Nevts, Areq, LossType, zmin, zmax, log10(Emin), log10(Emax), Seed);
  Output* myout = new Output(FileOut);
  cout << "** Output file is: " << FileOut << endl;
  cout << endl;
  cout << "** Events to be generated: " << Nevts << endl;


  
  //...... Loop over events ........
  
  for(event = 0; event<Nevts; event++) { 
    int branchxev = 0;
    start = end;
    cout << "** Event " << event << " ****************************************" << endl;
    //cout << "===> Seed: " <<   myRandom->GetSeed() << endl;
    const double Zini = myRandom->Uniform(zmin, zmax);
    const double Eini = pow(10,  myRandom->Uniform(log10(Emin), log10(Emax)));

    Particle *primary;

    if (proton) primary = new Particle(1, 1, 0, Eini, Zini, eProton);
    else if (random) primary = model->GetRandomParticle(Eini, Zini);
    else primary = new Particle(Areq, Zreq, 0, Eini, Zini, eNucleus);
    cout << "==> Primary nucleus (A,Z,E,z): " << primary->GetMassNum() << " " 
	 << primary->GetCharge() << " " << primary->GetEprod() << " " 
	 << primary->GetZprod() << endl;
    vector<Particle> stack;
    stack.push_back(*primary);
    while ( stack.size() > 0 ) {
      stack = Follow(stack, model, myout);
      branchxev++;
      //cout << "---------------------------- size of stack " << stack.size() << endl;
    }

    end = clock();
    
    cpu_time_used = ((double) (end - start)) / CLOCKS_PER_SEC;
    cout<<"** event ended - elapsed time (s) "<<cpu_time_used<<endl;
    unsigned long int CurrentSeed = (unsigned long int)myRandom->GetSeed();
    myout->Stream(cpu_time_used, branchxev, CurrentSeed); 
  }
 
  myout->Finish();

  printf("\n");
  printf("  ++++  Succesfully processed %7d events ++++ \n",Nevts);
  printf(" --------------------------------------------------\n"); 
  PrintSummary(Areq);
  
  return 0;
} 



//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

vector<Particle> Follow(vector<Particle> mystack, NucModel* mymodel, Output* myout) 
{
  
  Particle it = *(mystack.begin());
  mystack.erase(mystack.begin());
  if(it.GetType() == eProton) {
    mymodel->PropagateProton(&it);
  }
  else {
    vector<Particle> products = mymodel->PropagateNucleus(&it);
    if(products.size()>=1) { 
      for(unsigned int k=0; k<products.size(); k++) {
	mystack.push_back(products[k]);
      }
    }
  }
  vector<Particle> newstack = mystack;
  myout->Stream(event, &it);
  //cout << " it.GetZint() = " << it.GetZint() << " it.GetMassNum() = " << it.GetMassNum() << endl;  
  if(it.GetZint()==0) {
    NatEarth[it.GetMassNum()-1]++;
  }
  return newstack;

}



//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool GetOptions(int argc, char **argv)
{
  int c;
  ostringstream help;
  bool PrintHelp = false;
  help << "\n SimProp -s [random seed, D=65539] -N [number of events, D=100]\n"
       << "         -L [0=CMB only, 1=CMB+IR, 2=CMB+IR(power law), D=1]\n"
       << "         -A [A or 0 for all, D=56]\n"
       << "         -e [min nucleon energy log10(emin/eV), D=17] -E [max nucleon energy log10(emax/eV), D=21]\n"
       << "         -z [min redshift, D=0] -Z [max redshift, D=1]\n";
  
  while ((c = getopt (argc, argv, "s:N:L:A:e:E:z:Z:h")) != -1)
    
    switch (c)
      {
      case 's':
	Seed=atoi(optarg);
	break;
      case 'N':
	Nevts=atoi(optarg);
	break;
      case 'L':
	LossType=atoi(optarg);
	break;
      case 'A':
	Areq=atoi(optarg);
	break;
      case 'e':
	Emin=pow(10,atof(optarg));
	break;
      case 'E':
	Emax=pow(10,atof(optarg));
	break;
      case 'z':
	zmin=atof(optarg);
	break;
      case 'Z':
	zmax=atof(optarg);
	break;
      case 'h':
	PrintHelp=true;
	break;
      }
  

  if(PrintHelp) {
    cout << help.str() << endl;
    return false;
  }

  cout << "\n          >>  SimProp v" << VERSION << "r" << REVISION << "  <<\n\n";
  cout <<"------------------------------------------------------------" << endl;
  cout << "  Authors: D.Boncioli, A.F.Grillo, S.Petrera & F.Salamida" << endl; 
  cout <<"------------------------------------------------------------" << endl;
  cout << "  random seed      : " << Seed << "\n" 
       << "  events           : " << Nevts << "\n"
       << "  A                : " << Areq << "\n"
       << "  Loss type        : " << LossType << "\n"
       << "  Emin             : " << Emin << " eV\n"
       << "  Emax             : " << Emax << " eV\n"
       << "  Zmin             : " << zmin << "\n"
       << "  Zmax             : " << zmax << "\n";
  
  
  return true;
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool PrintSummary(int A)
{
 
  printf("       Nucleus Type   Number reaching Earth\n");
  printf(" --------------------------------------------------\n"); 
  int Ntot = 0;
  for (int i=0; i<A; i++) {
    Ntot += NatEarth[i];
    if(NatEarth[i]>0) 
      printf("      %8d     %14d\n",i+1,NatEarth[i]);
      //cout << "Nucleus n." << i+1 << " --> " << NatEarth[i] << endl;
  }
  printf(" --------------------------------------------------\n"); 
  if(Ntot>0)   printf("         Total     %14d\n\n",Ntot);
  return true;
}
